# -*- coding: utf-8 -*-
"""
Created on Sun Jun 21 16:55:22 2020

@author: Ian
"""


import dateutil as dateutil
import numpy as np
from datetime import datetime

#%%

def make_time_bounds_from_ds64(rec_avg_end, output_freq_code):
    """

    Given a datetime64 object (rec_avg_end) representing the 'end' of an 
    averaging time period (usually derived from the mitgcm file's timestep)
    and an output frequency code
    (AVG_MON, AVG_DAY, AVG_WEEK, or AVG_YEAR), create a time_bounds array
    with two datetime64 variables, one for the averaging period start, one
    for the averaging period end.  Also find the middle time between the
    two..
    
    Parameters
    ----------    
    rec_avg_end : numpy.datetime64 
        the time at the end of an averaging period

    output_freq_code : str 
        code indicating the time period of the averaging period
        - AVG_DAY, AVG_MON, AVG_WEEK, or AVG_YEAR


    Returns
    -------
    time_bnds : numpy.array(dtype=numpy.datetime64)
        a datetime64 array with the start and end time of the averaging periods
    
    center_times : numpy.datetime64
        the 'center' of the averaging period
    
    """ 
    
    if  output_freq_code in ('AVG_MON','AVG_DAY','AVG_WEEK','AVG_YEAR'):
        rec_year, rec_mon, rec_day, \
        rec_hour, rec_min, rec_sec = \
            extract_yyyy_mm_dd_hh_mm_ss_from_datetime64(rec_avg_end)
        
        
        rec_avg_end_as_dt = datetime(rec_year, rec_mon, 
                                     rec_day, rec_hour,
                                     rec_min, rec_sec)
        
        if output_freq_code     == 'AVG_MON':
            rec_avg_start =  rec_avg_end_as_dt - \
                dateutil.relativedelta.relativedelta(months=1)    
        elif output_freq_code   == 'AVG_DAY':
            rec_avg_start =  rec_avg_end_as_dt - \
                dateutil.relativedelta.relativedelta(days=1)  
        elif output_freq_code   == 'AVG_WEEK':
            rec_avg_start =  rec_avg_end_as_dt - \
                dateutil.relativedelta.relativedelta(weeks=1)  
        elif output_freq_code   == 'AVG_YEAR':
            rec_avg_start =  rec_avg_end_as_dt - \
                dateutil.relativedelta.relativedelta(years=1)    

        rec_avg_start =  np.datetime64(rec_avg_start)
        
        rec_avg_delta = rec_avg_end - rec_avg_start
        rec_avg_middle = rec_avg_start + rec_avg_delta/2
        #print rec_avg_end, rec_avg_start, rec_avg_middle
        
        rec_time_bnds = np.array([rec_avg_start, rec_avg_end])
        
        return rec_time_bnds, rec_avg_middle
    
    else:
        print ('output_freq_code must be: AVG_MON, AVG_DAY, AVG_WEEK, OR AVG_YEAR')
        print ('you provided ' + str(output_freq_code))
        return [],[]   
    
#%%
def extract_yyyy_mm_dd_hh_mm_ss_from_datetime64(dt64):
    """

    Extract separate fields for year, monday, day, hour, min, sec from
    a datetime64 object
    
    Parameters
    ----------    
    dt64 : numpy.datetime64 
        a datetime64 object

    Returns 
    -------    
    year, mon, day, hh, mm, ss : int

    """
    
    s = str(dt64)
    year = int(s[0:4])
    mon = int(s[5:7])
    day = int(s[8:10])
    hh = int(s[11:13])
    mm = int(s[14:16])
    ss = int(s[17:18])
    
    #print year, mon, day, hh, mm, ss
    return year,mon,day,hh,mm,ss 



#%%
def months2days(nmon=288, baseyear=1992, basemon=1):
    """ 

    This routine converts the mid-month time to days from January 1st of a particular year.
    
    Parameters
    ----------
    nmon : dtype=integer
        number of months 

    baseyear : dtype=integer
        year of time of origin
        
    basemon : dtype=integer
        month of time of origin 
    
    Returns
    -------
        time_days : ndarray
            the middle time of each month in days from Jan 1 baseyear (numpy array [nmon], dtype=double)
        
        time_days_bnds : ndarray
            time bounds (numpy array [nmon, 2], dtype=double)

        ansi_date : ndarray
            array of ANSI date strings

    """
    
    time_days_bnds = np.zeros([nmon,2])
    time_1stdayofmon = np.zeros([nmon+1])
    
    basetime = datetime.datetime(baseyear, basemon, 1, 0, 0, 0)

    for mon in range(nmon+1):
        #monfrombasemon is how many months fron basemon
        monfrombasemon=basemon+mon-1
        yrtmp = monfrombasemon//12+baseyear
        montmp = monfrombasemon%12+1
        tmpdate = datetime.datetime(yrtmp,montmp,1,0,0,0)-basetime
        time_1stdayofmon[mon] = tmpdate.days
    #time bounds are the 1st day of each month.
    time_days_bnds[:,0]= time_1stdayofmon[0:nmon]
    time_days_bnds[:,1]= time_1stdayofmon[1:nmon+1]
    #center time of each month is the mean of the time bounds.    
    time_days = np.mean(time_days_bnds,axis=1)

    ansi_datetmp = np.array([basetime + datetime.timedelta(days=time_days[i]) for i in xrange(nmon)])
    ansi_date = [str.replace(ansi_datetmp[i].isoformat(),'T',' ') for i in range(nmon)]

    return time_days, time_days_bnds, ansi_date

#%%


from .date_time import make_time_bounds_from_ds64
from .date_time import extract_yyyy_mm_dd_hh_mm_ss_from_datetime64
from .date_time import months2days

from .llc_array_conversion  import llc_tiles_to_compact
from .llc_array_conversion  import llc_tiles_to_faces
from .llc_array_conversion  import llc_faces_to_compact

from .records import make_empty_record
from .records import save_to_disk

from .mapping import find_mappings_from_source_to_target
from .mapping import transform_to_target_grid

from .geometry import area_of_latlon_grid_cell
from .geometry import area_of_latlon_grid

__all__ = ['date_time', 'llc_array_conversion', 'records', 'mapping', 'geometry']



# -*- coding: utf-8 -*-
"""
Created on Sun Jun 21 10:10:13 2020

@author: Ian
"""

import numpy as np




# %%
def llc_tiles_to_compact(data_tiles, less_output = False):
    """

    Converts a numpy binary array in the 'compact' format of the 
    lat-lon-cap (LLC) grids and converts it to the '13 tiles' format
    of the LLC grids.  

    Parameters
    ----------
    data_tiles : ndarray
        a numpy array organized by, at most, 
        13 tiles x nl x nk x llc x llc
    
        where dimensions 'nl' and 'nk' are optional.
        
    less_output : boolean, optional, default False
        A debugging flag.  False = less debugging output
        
    Returns
    -------
    data_compact : ndarray
        a numpy array of dimension nl x nk x 13*llc x llc 

    Note
    ----
    If dimensions nl or nk are singular, they are not included 
    as dimensions in data_compact

    """
   
    data_faces   = llc_tiles_to_faces(data_tiles, less_output=less_output)
    data_compact = llc_faces_to_compact(data_faces, less_output=less_output)
        
    return data_compact



def llc_tiles_to_faces(data_tiles, less_output=False):
    """

    Converts an array of 13 'tiles' from the lat-lon-cap grid
    and rearranges them to 5 faces.  Faces 1,2,4, and 5 are approximately 
    lat-lon while face 3 is the Arctic 'cap' 

    Parameters
    ----------
    data_tiles : 
        An array of dimension 13 x nl x nk x llc x llc
        
    If dimensions nl or nk are singular, they are not included 
        as dimensions of data_tiles    

    less_output : boolean
        A debugging flag.  False = less debugging output
        Default: False
    
    Returns
    -------
    F : dict
        a dictionary containing the five lat-lon-cap faces
        
        F[n] is a numpy array of face n, n in [1..5]

        dimensions of each 2D slice of F
        
        - f1,f2: 3*llc x llc
        -    f3: llc x llc
        - f4,f5: llc x 3*llc  
    
    """
    
    # ascertain how many dimensions are in the faces (minimum 3, maximum 5)
    dims = data_tiles.shape
    num_dims = len(dims)
    
    # the final dimension is always length llc
    llc = dims[-1]

    # tiles is always just before (y,x) dims
    num_tiles = dims[-3]

    if less_output == False:
        print('num tiles, ', num_tiles)

    if num_dims == 3: # we have a 13 2D slices (tile, y, x)
        f1 = np.zeros((3*llc, llc))
        f2 = np.zeros((3*llc, llc))
        f3 = np.zeros((llc, llc))
        f4 = np.zeros((llc, 3*llc))
        f5 = np.zeros((llc, 3*llc))

    elif num_dims == 4: # 13 3D slices (time or depth, tile, y, x)

        nk = dims[0]
        
        f1 = np.zeros((nk, 3*llc, llc))
        f2 = np.zeros((nk, 3*llc, llc))
        f3 = np.zeros((nk, llc, llc))
        f4 = np.zeros((nk, llc, 3*llc))
        f5 = np.zeros((nk, llc, 3*llc))

    elif num_dims == 5: # 4D slice (time or depth, time or depth, tile, y, x)
        nl = dims[0]
        nk = dims[1]

        f1 = np.zeros((nl,nk, 3*llc, llc))
        f2 = np.zeros((nl,nk, 3*llc, llc))
        f3 = np.zeros((nl,nk, llc, llc))
        f4 = np.zeros((nl,nk, llc, 3*llc))
        f5 = np.zeros((nl,nk, llc, 3*llc))

    else:
        print ('can only handle tiles that have 2, 3, or 4 dimensions!')
        return []

    # Map data on the tiles to the faces

    # 2D slices on 13 tiles
    if num_dims == 3: 
        
        f1[llc*0:llc*1,:] = data_tiles[0,:]

        f1[llc*1:llc*2,:] = data_tiles[1,:]
        f1[llc*2:,:]      = data_tiles[2,:]

        f2[llc*0:llc*1,:] = data_tiles[3,:]
        f2[llc*1:llc*2,:] = data_tiles[4,:]
        f2[llc*2:,:]      = data_tiles[5,:]
        
        f3                = data_tiles[6,:]

        f4[:,llc*0:llc*1] = data_tiles[7,:]
        f4[:,llc*1:llc*2] = data_tiles[8,:]
        f4[:,llc*2:]      = data_tiles[9,:]
        
        f5[:,llc*0:llc*1] = data_tiles[10,:]
        f5[:,llc*1:llc*2] = data_tiles[11,:]
        f5[:,llc*2:]      = data_tiles[12,:]

    # 3D slices on 13 tiles
    elif num_dims == 4: 

        for k in range(nk):
            f1[k,llc*0:llc*1,:] = data_tiles[k,0,:]

            f1[k,llc*1:llc*2,:] = data_tiles[k,1,:]
            f1[k,llc*2:,:]      = data_tiles[k,2,:]

            f2[k,llc*0:llc*1,:] = data_tiles[k,3,:]
            f2[k,llc*1:llc*2,:] = data_tiles[k,4,:]
            f2[k,llc*2:,:]      = data_tiles[k,5,:]
            
            f3[k,:]             = data_tiles[k,6,:]

            f4[k,:,llc*0:llc*1] = data_tiles[k,7,:]
            f4[k,:,llc*1:llc*2] = data_tiles[k,8,:]
            f4[k,:,llc*2:]      = data_tiles[k,9,:]
            
            f5[k,:,llc*0:llc*1] = data_tiles[k,10,:]
            f5[k,:,llc*1:llc*2] = data_tiles[k,11,:]
            f5[k,:,llc*2:]      = data_tiles[k,12,:]

    # 4D slices on 13 tiles
    elif num_dims == 5: 
        for l in range(nl):
            for k in range(nk):
                f1[l,k,llc*0:llc*1,:] = data_tiles[l,k,0,:]

                f1[l,k,llc*1:llc*2,:] = data_tiles[l,k,1,:]
                f1[l,k,llc*2:,:]      = data_tiles[l,k,2,:]

                f2[l,k,llc*0:llc*1,:] = data_tiles[l,k,3,:]
                f2[l,k,llc*1:llc*2,:] = data_tiles[l,k,4,:]
                f2[l,k,llc*2:,:]      = data_tiles[l,k,5,:]
                
                f3[l,k,:]             = data_tiles[l,k,6,:]

                f4[l,k,:,llc*0:llc*1] = data_tiles[l,k,7,:]
                f4[l,k,:,llc*1:llc*2] = data_tiles[l,k,8,:]
                f4[l,k,:,llc*2:]      = data_tiles[l,k,9,:]
                
                f5[l,k,:,llc*0:llc*1] = data_tiles[l,k,10,:]
                f5[l,k,:,llc*1:llc*2] = data_tiles[l,k,11,:]
                f5[l,k,:,llc*2:]      = data_tiles[l,k,12,:]

    # Build the F dictionary
    F = {}
    F[1] = f1
    F[2] = f2
    F[3] = f3
    F[4] = f4
    F[5] = f5
    
    return F


def llc_faces_to_compact(F, less_output=True):
    """
    
    Converts a dictionary containing five 'faces' of the lat-lon-cap grid
    and rearranges it to the 'compact' llc format.


    Parameters
    ----------
    F : dict
        a dictionary containing the five lat-lon-cap faces
        
        F[n] is a numpy array of face n, n in [1..5]

        dimensions of each 2D slice of F

        - f1,f2: 3*llc x llc
        -    f3: llc x llc
        - f4,f5: llc x 3*llc  
    
    less_output : boolean, optional, default False
        A debugging flag.  False = less debugging output
        
    Returns
    -------
    data_compact : ndarray
        an array of dimension nl x nk x nj x ni 
        
        F is in the llc compact format.

    Note
    ----
    If dimensions nl or nk are singular, they are not included 
    as dimensions of data_compact

    """

    # pull the individual faces out of the F dictionary
    f1 = F[1]
    f2 = F[2]
    f3 = F[3]
    f4 = F[4]
    f5 = F[5]

    # ascertain how many dimensions are in the faces (minimum 2, maximum 4)
    dims = f3.shape
    num_dims = len(dims)

    # the final dimension is always the llc # 
    llc = dims[-1]

    # initialize the 'data_compact' array
    if num_dims == 2: # we have a 2D slice (x,y)
        data_compact = np.zeros((13*llc, llc))

    elif num_dims == 3: # 3D slice (x, y, time or depth)
        nk = dims[0]
        data_compact = np.zeros((nk, 13*llc, llc))

    elif num_dims == 4: # 4D slice (x,y,time and depth)
        nl = dims[0]
        nk = dims[1]
        data_compact = np.zeros((nl, nk, 13*llc, llc))
    else:
        print ('can only handle faces that have 2, 3, or 4 dimensions!')
        return []

    if less_output == False:
        print ('shape of face 3 ', f3.shape)

    if num_dims == 2:
        
        data_compact[:3*llc,:]      = f1
        data_compact[3*llc:6*llc,:] = f2
        data_compact[6*llc:7*llc,:] = f3
        
        for f in range(8,11):
            i1 = np.arange(0, llc)+(f-8)*llc
            i2 = np.arange(0,3*llc,3) + 7*llc + f -8
            data_compact[i2,:] = f4[:,i1]

        for f in range(11,14):
            i1 = np.arange(0, llc)+(f-11)*llc
            i2 = np.arange(0,3*llc,3) + 10*llc + f -11
            data_compact[i2,:] = f5[:,i1]
    
    elif num_dims == 3:
        # loop through k indicies
        print ('size of data compact ', data_compact.shape)
       
        for k in range(nk):
            data_compact[k,:3*llc,:]      = f1[k,:]
            data_compact[k,3*llc:6*llc,:] = f2[k,:]
            data_compact[k,6*llc:7*llc,:] = f3[k,:]

            # if someone could explain why I have to transpose
            # f4 and f5 when num_dims =3 or 4 I would be so grateful.
            # Just could not figure this out.  Transposing works but why!?
            for f in range(8,11):
                i1 = np.arange(0, llc)+(f-8)*llc
                i2 = np.arange(0,3*llc,3) + 7*llc + f - 8

                data_compact[k,i2,:] = f4[k,0:llc,i1].T

            for f in range(11,14):
                i1 = np.arange(0, llc)+(f-11)*llc
                i2 = np.arange(0,3*llc,3) + 10*llc + f -11
                data_compact[k,i2,:] = f5[k,:,i1].T

    elif num_dims == 4:
        # loop through l and k indices
        for l in range(nl):
            for k in range(nk):
                data_compact[l,k,:3*llc,:]      = f1[l,k,:]
                data_compact[l,k,3*llc:6*llc,:] = f2[l,k,:]
                data_compact[l,k,6*llc:7*llc,:] = f3[l,k,:]
                
                for f in range(8,11):
                    i1 = np.arange(0, llc)+(f-8)*llc
                    i2 = np.arange(0,3*llc,3) + 7*llc + f -8
                    data_compact[l,k,i2,:]      = f4[l,k,:,i1].T

                for f in range(11,14):
                    i1 = np.arange(0, llc)+(f-11)*llc
                    i2 = np.arange(0,3*llc,3) + 10*llc + f -11
                    data_compact[l,k,i2,:]      = f5[l,k,:,i1].T


    if less_output == False:
        print ('shape of data_compact ', data_compact.shape)

    return data_compact


# -*- coding: utf-8 -*-
"""
Created on Sun Jun 21 19:28:40 2020

@author: Ian
"""
import numpy as np
import pyresample as pr


#%%


#%%
def transform_to_target_grid(source_indices_within_target_radius_i,
                             num_source_indices_within_target_radius_i,
                             nearest_source_index_to_target_index_i,
                             source_field, target_grid_shape,\
                             operation='mean', allow_nearest_neighbor=True):
                            
    # source_indices_within_target_radius_i
    # num_source_indices_within_target_radius_i
    # nearest_source_index_to_target_index_i
    # source field: 2D field
    # target_grid_shape : shape of target grid array (2D)
    # operation : mean or median
    

    source_field_r = source_field.ravel()
    
    # define array that will contain source_field mapped to target_grid
    source_on_target_grid = np.zeros((target_grid_shape))*np.nan
    
    # get a 1D version of source_on_target_grid
    tmp_r = source_on_target_grid.ravel()
 
    # loop through every target grid point
    for i in range(len(tmp_r)):
        
        # if the number of source_field points at target grid cell i > 0
        if num_source_indices_within_target_radius_i[i] > 0:
            
            # average these values
            if operation == 'mean':
                tmp_r[i] = \
                    np.mean(source_field_r[source_indices_within_target_radius_i[i]])
            # median of these values 
            elif operation == 'median':
                tmp_r[i] = \
                    np.median(source_field_r[source_indices_within_target_radius_i[i]])
      
        # if there aren't any source grid cells within the target grid cell
        # search radius then we would use the nearest neighbor, if available
        elif i in nearest_source_index_to_target_index_i.keys():
            # just because we have a nearest neighbor doesn't mean use it
            if allow_nearest_neighbor == True:
                tmp_r[i] = source_field_r[nearest_source_index_to_target_index_i[i]]
                           
     
    return source_on_target_grid



#%%
def find_mappings_from_source_to_target(source_grid, target_grid,\
                                        target_grid_radius, \
                                        source_grid_min_L, source_grid_max_L, \
                                        neighbours = 100):
                                    
    # source grid, target_grid : area or grid defintion objects from pyresample
    
    # target_grid_radius       : a vector indicating the radius of each
    #                            target grid cell (m)
    
    # source_grid_min_l, source_grid_max_L : min and max distances 
    #                            between adjacent source grid cells (m)
    
    # neighbours     : Specifies number of neighbours to look for when getting
    #                  the neighbour info of a cell using pyresample.
    #                  Default is 100 to limit memory usage.
    #                  Value given must be a whole number greater than 0
    
    # # of element of the source and target grids
    len_source_grid = source_grid.size
    len_target_grid = target_grid.size
    
    # the maximum radius of the target grid
    max_target_grid_radius = np.nanmax(target_grid_radius)
    
    # the maximum number of neighbors to consider when doing the bin averaging
    # assuming that we have the largets target grid radius and the smallest
    # source grid length. (upper bound)
    
    # the ceiling is used to ensure the result is a whole number > 0
    neighbours_upper_bound = np.ceil((max_target_grid_radius*2/source_grid_min_L)**2)
    
    # compare provided and upper_bound value for neighbours.
    # limit neighbours to the upper_bound if the supplied neighbours value is larger
    # since you dont need more neighbours than exists within a cell.
    if neighbours > neighbours_upper_bound:
        print('using more neighbours than upper bound.  limiting to the upper bound ' \
              f'of {int(neighbours_upper_bound)} neighbours')
        neighbours = neighbours_upper_bound
        
    # make sure neighbours is an int for pyresample
    # neighbours_upper_bound is float, and user input can be float
    neighbours = int(neighbours)
    
    ## FIRST FIND THE SET OF SOURCE GRID CELLS THAT FALL WITHIN THE SERACH
    ## RADIUS OF EACH TARGET GRID CELL
    
    # "target_grid_radius" is the half of the distance between 
    # target grid cells.  No need to search for source
    # grid points more than halfway to the next target grid cell.
    
    # the get_neighbour_info returned from pyresample is quite useful.  
    # Ax[2] is the matrix of 
    # closest data grid points for each model grid point
    # Ax[3] is the actual distance in meters
    # also cool is that Ax[3] is sorted, first column is closest, last column
    # is furthest.
    # for some reason the radius of influence has to be in an int.
    
    Ax_max_target_grid_r = \
        pr.kd_tree.get_neighbour_info(source_grid, \
                                      target_grid, \
                                      radius_of_influence=int(max_target_grid_radius),\
                                      neighbours=neighbours)
    
    # define a dictionary, which will contain the list of SOURCE grid cells
    # that are within the search radius of each TARGET grid cell 
    source_indices_within_target_radius_i = dict()
    
    # define a vector which is a COUNT of the # of SOURCE grid cells
    # that are within the search radius of each TARGET grid cell 
    num_source_indices_within_target_radius_i =\
        np.zeros((target_grid_radius.shape))
    

    ## SECOND FIND THE SINGLE SOURCE GRID CELL THAT IS CLOSEST TO EACH 
    ## TARGET GRID CELL, BUT ONLY SEARCH AS FAR AS SOURCE_GRID_MAX_L
    
    # the kd_tree can also find the sigle nearest neighbor within the
    # radius 'source_grid_max_L'.  This second search is needed because sometimes
    # the source grid is finer than the target grid and therefore we may 
    # end up in a situation where none of the centers of the SOURCE grid
    # fall within the small centers of the TARGET grid.
    # we'll look for the nearest SOURCE grid cell within 'source_grid_max_L'       
    
    Ax_nearest_within_source_grid_max_L = \
        pr.kd_tree.get_neighbour_info(source_grid, target_grid, \
                                      radius_of_influence=int(source_grid_max_L), \
                                      neighbours=1)
    
    # define a vector that will store the index of the source grid closest to
    # the target grid within the search radius 'source_grid_max_L'
    nearest_source_index_to_target_index_i = dict()


    # >> a list of i's to print debug statements for
    debug_is =  np.linspace(0,len_target_grid,20).astype(int)
    
    # loop through every model grid cell, pull out the SOURCE grid cells
    # that fall within the target grid radius and stick into the 
    # 'source_indices_within_target_radius_i' dictionary
    # and then do the same for the nearest neighbour business.
    
    current_valid_target_i = 0
    
    for i in range(len_target_grid):
        
        if Ax_nearest_within_source_grid_max_L[1][i] == True:
            
            # Ax[2][i,:] are the closest source grid indices 
            #            for target grid cell i
            # data_within_search_radius[i,:] is the T/F array for which
            #            of the closest 'neighbours' source grid indices are within
            #            the radius of this target grid cell i
            # -- so we're pulling out just those source grid indices
            #    that fall within the target grid cell radius
            
            ## FIRST RECORD THE SOURCE POINTS THAT FALL WITHIN TARGET_GRID_RADIUS
            dist_from_src_to_target = Ax_max_target_grid_r[3][current_valid_target_i]
            
            dist_within_target_r = dist_from_src_to_target <= target_grid_radius[i]
           
            src_indicies_here = Ax_max_target_grid_r[2][current_valid_target_i,:]            
        
            source_indices_within_target_radius_i[i] = \
                src_indicies_here[dist_within_target_r == True]
               
            # count the # source indices here 
            num_source_indices_within_target_radius_i[i] = \
                int(len(source_indices_within_target_radius_i[i] ))
            
            
        
            ## NOW RECORD THE NEAREST NEIGHBOR POINT WIHTIN SOURCE_GRID_MAX_L
            # when there is no source index within the search radius then
            # the 'get neighbour info' routine returns a dummy value of 
            # the length of the source grid.  so we test to see if that's the
            # value that was returned.  If not, then we are good to go.
            if Ax_nearest_within_source_grid_max_L[2][current_valid_target_i]\
                < len_source_grid:
                nearest_source_index_to_target_index_i[i] =\
                    Ax_nearest_within_source_grid_max_L[2][current_valid_target_i]
        
            # increment this little bastard            
            current_valid_target_i += 1
        
        # print progress.  always nice 
        if i in debug_is:
            print(str(int(i/len_target_grid*100)) + ' %')

    return source_indices_within_target_radius_i,\
           num_source_indices_within_target_radius_i,\
           nearest_source_index_to_target_index_i





# -*- coding: utf-8 -*-
"""
Created on Sun Jun 21 17:24:13 2020

@author: Ian
"""
import xarray as xr
import numpy as np
from pathlib import Path
from .llc_array_conversion  import llc_tiles_to_compact

#%%
def make_empty_record(standard_name, long_name, units,\
                      record_date,\
                      model_grid, model_grid_type,\
                      array_precision):
    
    # make an empty data array to hold the interpolated 2D field
    # all values are nans. 
    # dimensions are the same as model_grid.XC
    data_DA =xr.DataArray(np.ones(np.shape(model_grid.XC.values),\
                                  dtype=array_precision), \
                            dims=model_grid.XC.dims)*np.nan

    data_DA=data_DA.assign_coords(time = np.datetime64(record_date,'ns'))
    data_DA=data_DA.expand_dims(dim = 'time', axis=0)
    
    # add start and end time records. default is same value as record date
    data_DA = data_DA.assign_coords(time_start = ('time',data_DA.time.copy(deep=True)))
    data_DA = data_DA.assign_coords(time_end = ('time',data_DA.time.copy(deep=True)))

    for dim in model_grid.XC.dims:
        data_DA=data_DA.assign_coords({dim : model_grid[dim]})

    
    # create XC and YC coordinates of empty record from XC YC of model_grid
    if model_grid_type == 'llc':
        # llc grid has 'tile dimension'
        data_DA=data_DA.assign_coords({'XC': (('tile','j','i'), model_grid.XC)})
        data_DA=data_DA.assign_coords({'YC': (('tile','j','i'), model_grid.YC)})
 
    # some grids only have j i dimensions
    elif model_grid_type == 'latlon':
        data_DA=data_DA.assign_coords({'XC': (('j','i'), model_grid.XC)})
        data_DA=data_DA.assign_coords({'YC': (('j','i'), model_grid.YC)})
     
    else: 
        print('invalid grid type!')
        return []
     
    # copy over the attributes from XC and YC to the dataArray
    data_DA.XC.attrs = model_grid.XC.attrs
    data_DA.YC.attrs = model_grid.YC.attrs
 
    # add some metadata
    data_DA.attrs = []
    data_DA.attrs['interpolated_grid'] = model_grid.title
    data_DA.attrs['model_grid_type'] = model_grid_type
    data_DA.attrs['long_name'] = long_name
    data_DA.attrs['standard_name'] = standard_name
    data_DA.attrs['units'] = units

    return data_DA

#%%

def save_to_disk(data_DA, \
                 output_filename, \
                 binary_fill_value, netcdf_fill_value,\
                 netcdf_output_dir,binary_output_dir, binary_output_dtype,\
                 model_grid_type) :  
    
    # define binary file output filetype    
    dt_out = np.dtype(binary_output_dtype)

    # create directories
    netcdf_output_dir.mkdir(exist_ok=True)
    binary_output_dir.mkdir(exist_ok=True)

    
    netcdf_output_filename = netcdf_output_dir / Path(output_filename + '.nc')
    binary_output_filename = binary_output_dir / output_filename
        
    # replace nans with the binary fill value (something like -9999)
    tmp_fields = np.where(np.isnan(data_DA.values), \
                          binary_fill_value, data_DA.values)


    ### SAVE FLAT BINARY
    # loop through each record of the year, save binary fields one at a time
    # appending each record as we go
    fd1 = open(str(binary_output_filename),'wb')
    fd1 = open(str(binary_output_filename),'ab')

    for i in range(len(data_DA.time)):
        print ('saving record: ', str(i))
        
        # if we have an llc grid, then we have to reform to compact
        if model_grid_type == 'llc':
            tmp_field = llc_tiles_to_compact(tmp_fields[i,:], less_output=True)
            
        # otherwise assume grid is x,y (2 dimensions)
        elif model_grid_type == 'latlon':
            tmp_field  = tmp_fields[i,:]
            
        else:
            print('unknown model grid type!')
            tmp_field = []
            return []
           
        # make sure we have something to save...
        if len(tmp_field) > 0:
            # if this is the first record, create new binary file
                tmp_field.astype(dt_out).tofile(fd1)
                
    # close the file at the end of the operation
    fd1.close()
                
    ### SAVE NETCDF
    # replace the binary fill value (-9999) with the netcdf fill value
    # which is much more interesting
    
    # replace nans with the binary fill value (something like -9999)
    data_DA.values = \
        np.where(np.isnan(data_DA.values),\
                 netcdf_fill_value, data_DA.values)
    
    encoding_each = {'zlib':True, \
                     'complevel':9,\
                     'fletcher32':True,\
                     '_FillValue':netcdf_fill_value}
    
    data_DS = data_DA.to_dataset()
    
    encoding = {var:encoding_each for var in data_DS.data_vars}
  
    # the actual saving (so easy with xarray!)    
    data_DS.to_netcdf(netcdf_output_filename,  encoding=encoding)
    data_DS.close()


# -*- coding: utf-8 -*-
"""
Created on Sun Jun 21 19:28:40 2020

@author: Ian
"""
import numpy as np


#%%


def area_of_latlon_grid_cell(lon0, lon1, lat0, lat1):
    #https://gis.stackexchange.com/questions/29734/how-to-calculate-area-of-1-x-1-degree-cells-in-a-raster
    

    #It is a consequence of a theorem of Archimedes (c. 287-212 BCE) that 
    #for a spherical model of the earth, the area of a cell spanning 
    #longitudes l0 to l1 (l1 > l0) and latitudes f0 to f1 (f1 > f0) equals
    
    #(sin(f1) - sin(f0)) * (l1 - l0) * R^2
    
    #where
    #
    #    l0 and l1 are expressed in radians (not degrees or whatever).
    #    l1 - l0 is calculated modulo 2*pi (e.g., -179 - 181 = 2 degrees, not -362 degrees).
    #
    #    R is the authalic Earth radius, almost exactly 6371 km.
    #    (sin(f1) - sin(f0)) * (l1 - l0) * R^2
        
    R_Earth = 6371.0088e3 # m
    
    A = (np.sin(np.deg2rad(lat1)) - np.sin(np.deg2rad(lat0))) * \
        (np.deg2rad(lon1) - np.deg2rad(lon0)) * \
        R_Earth **2

    return A


def area_of_latlon_grid(lon0, lon1, lat0, lat1, dx, dy):
    # Calculates area of a latlon grid with edges lon0 and lon1
    # lat0 and lat1 with grid spacing of dx and dy
    
    # lons and lats are in degrees
    # dx and y are in degrees
    
    # resulting array has columns of lon, rows of lat.
    
    # Using -180, 180, -90, 90 we get total area of 510065.88 x10^6 km^2
    # using     R_Earth = 6371.0088e3 # m

    num_grid_cells_x = int((lon1-lon0)/dx)
    num_grid_cells_y = int((lat1-lat0)/dy)
    
    lons_grid_cell_edges = np.linspace(lon0, lon1, num_grid_cells_x + 1)
    lats_grid_cell_edges = np.linspace(lat0, lat1, num_grid_cells_y + 1)

    A = np.zeros((num_grid_cells_y))
    
    print(lons_grid_cell_edges)
    print(lats_grid_cell_edges)
    
    for lat_i in range(num_grid_cells_y):
        A[lat_i] = area_of_latlon_grid_cell(lons_grid_cell_edges[0], lons_grid_cell_edges[1],\
         lats_grid_cell_edges[lat_i], lats_grid_cell_edges[lat_i+1])
        
    return np.tile(A, (num_grid_cells_x,1)).T





import setuptools

from distutils.core import setup

setup(
  name = 'ecco_cloud_utils',
  packages = ['ecco_cloud_utils'], # this must be the same as the name above
  version = '0.0.2',
  description = 'Estimating the Circulation and Climate of the Ocean (ECCO) - Cloud Package',
  author = 'Ian Fenty',
  author_email = 'ian.fenty@jpl.nasa.gov',
  url = 'https://github.com/ECCO-GROUP/ECCO-CLOUD/ecco-cloud-utils',
  keywords = ['ecco','mitgcm','estimate','circulation','climate'],
  include_package_data=True,
 # data_files=[('binary_data',['binary_data/basins.data', 'binary_data/basins.meta'])],
  install_requires=['datetime', 'python-dateutil', 'numpy','pyresample','xarray','pathlib'],
  download_url= 'http://github.com/ECCO-GROUP/ECCO-ACCESS/archive/0.0.2.zip',
  classifiers=[
      'Development Status :: 5 - Production/Stable',
      'Intended Audience :: Science/Research', 
      'License :: OSI Approved :: MIT License',
      'Natural Language :: English',
      'Programming Language :: Python',
      'Programming Language :: Python :: 3.7',
      'Topic :: Scientific/Engineering :: Physics'
  ]
)
